#!/usr/bin/python
##############################################################################
#
# Copyright (c) 2002, 2006 Nexedi SARL and Contributors. All Rights Reserved.
#                    Bartlomiej Gorny <bartek@erp5.pl>
#                    Jacek Medrzycki <jacek@erp5.pl>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################


import cStringIO
import glob
import os
import pdb
import random
import re
import socket
import threading
import time
import zipfile
import shutil

# for python-2.3
if getattr(threading, 'local', None) is None:
  from _threading_local import local
  threading.local = local


from SimpleXMLRPCServer import SimpleXMLRPCServer
from SocketServer import ThreadingMixIn

import config
import lib
from logger import Log
import oood_common
from pool import pool

# error codes
OK = 200
NOFUNC = 401
BAD_DOCUMENT = 402
TIMEOUT = 403
BAD_REQUEST = 404
BLOCKED = 501
POOL_EMPTY = 502
OUT_OF_SYNC = 503
UNKNOWN = 504

def formatTime(t):
  return time.strftime('%H:%M:%S', time.gmtime(t))

class MySerw(ThreadingMixIn, SimpleXMLRPCServer):

  def server_bind(self):
    self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    self.socket.bind(self.server_address)

# XXX shouldn't we make all threads daemonic?

class Dispatcher(object):
  """
    This is a replacement for serw.Procesor class
    Request handling procedure:
    # check if server is oood_has_blocked, return error if it is (means we fit into a very narrow time span)
    # check if pool size > 0, otherwise reinitialize
    # create local object threading.local for all flags and references
    # start semaphore_waiting thread
    # it the semaphore_waiting thread times out,
    #     mark all active threads as oood_has_blocked
    #     mark server as oood_has_blocked
    #     release all semaphores so that they can return
    #     reinitialize everything
    # acquire semaphore
    # if current thread is oood_has_blocked: return error message
    # get worker or trigger a reinitialization and return error message
    # start process_waiting thread, cross-referenced with current thread
    # if process_waiting thread times out, mark current thread as "timed_out", restart worker, release semaphore
    # before returning anything from any public function, check if current thread is not timed_out, if it is then return error message
    # try do to what is requested
    # produce and return the result
    # if there was any exception, then:
    # cancel process_waiting thread, start another
    # restart worker
    # try again
    # if there was any exception, then:
    # cancel process_waiting thread
    # restart worker
    # release semaphore
  """

  def __init__(self, offset_str=''):
    # create semaphore
    self.semaphore = threading.Semaphore(config.pool_size)
    # init flags
    self.oood_has_blocked = False
    # init locks
    self.emergency_restart_lock = threading.Lock()
    # initialize
    self.creation_time = int(time.time())
    self.counter = 0
    self.restart_count = -1
    self.offset_str=offset_str
    self.tmp_dir_name = config.tmp_dir+offset_str
    self.offset=0
    if offset_str != '':
      self.offset = int(offset_str)
    self.initialize()


  def initialize(self):
    # cancel all waiting timers
    self.start_time = int(time.time())
    self.restart_count += 1
    Log.info('initializing...')
    for thr in threading.enumerate():
      try:
        thr.cancel()
      except AttributeError:
        pass
    # acquire all semaphores (in non-blocking mode), so that new requests wait till we're done
    for i in range(config.pool_size):
      self.semaphore.acquire(False)
    # recreate pool and workers
    pool.initializePool(self.offset)
    # tell threads we are not blocked (if we are re-initializing)
    # XXX is the sequence below correct?
    for thr in threading.enumerate():
      thr.oood_has_blocked = False
    # reset flags
    self.oood_has_blocked = False
    # release all semaphores - now they can go ahead
    for i in range(config.pool_size):
      self.semaphore.release()
    # initialization cancelled all timers, so now we can safely release
    self.emergency_restart_lock.acquire(False)
    self.emergency_restart_lock.release()

  def _dispatch(self, funcname, a):
    """
      Exception handling wrapper for the real dispatcher
    """
    try:
      return self.do_dispatch(funcname, a)
    except Exception, e:
      Log.logException(e)
      return self.error(504, str(e)).asResponse()

  def do_dispatch(self, funcname, a):
    """
      General request handler
      XMLRPC protocol does not support named arguments
    """
    if funcname == 'trace':
      pdb.set_trace()
    if funcname == 'threadingStatus':
      return self.threadingStatus()
    if funcname == 'poolStatus':
      return self.poolStatus()
    # build a request object from what we got, check if it is ok
    try:
      request = oood_common.requestFactory(a)
    except oood_common.MalformedRequest, e:
      response = self.error(BAD_REQUEST, str(e))
      return response.asResponse()
    if self.oood_has_blocked: # means we are in the process of an emergency restart
      response = self.error(BLOCKED)
      return response.asResponse()
    # if pool has no workers, trigger reinitialization and return error
    # XXX this is not needed now since we alway run it with one worker
    #if pool.getSize() == 0:
      #emergency_thread = threading.Timer(1, self.emergencyRestart)
      #emergency_thread.start()
      #response = self.error(POOL_EMPTY)
      #return response.asResponse()
    this = threading.currentThread()
    # init thread flags
    this.oood_has_blocked = False
    this.timed_out = False
    local = threading.local()
    # start controlling thread to restart if we can't acquire semaphore
    # XXX this can cause problems - if the serwer gets too many requests, it will get emergency restart only because
    # it is overloaded - we will time out in waiting for semaphore, while the server is processing previous requests
    # the more so that we allow lots of waiting in pool.makeWorkerObject
    max_allowed_time = (config.instance_timeout + config.instance_load_time) * 3
    local.semaphore_waiting = threading.Timer(max_allowed_time, self.emergencyRestart)
    Log.debug('%s is waiting for semaphore', local.semaphore_waiting.getName())
    local.semaphore_waiting.start()
    self.semaphore.acquire() # blocking mode
    # ok, we have it
    Log.debug('we acquired semaphore')
    local.semaphore_waiting.cancel()
    # if we blocked out in the meantime, return error
    if this.oood_has_blocked:
      self.semaphore.release()
      response = self.error(BLOCKED)
      return response.asResponse()
    # get worker
    worker = pool.getWorkerObject()
    # if none available, we are out of sync so we have to return error and restart
    if not worker:
      emergency_thread = threading.Timer(1, self.emergencyRestart)
      emergency_thread.start()
      response = self.error(OUT_OF_SYNC)
      return response.asResponse()
    # start thread to timeout document processing
    local.process_waiting = threading.Timer(config.instance_timeout, self.instanceRestart, kwargs={'target':this, 'worker':worker})
    Log.debug('%s is waiting for process to complete', local.process_waiting.getName())
    local.process_waiting.start()
    # try do do what was requested
    try:
      try:
        try:
          data = self._process(worker, funcname, request)
        except NotImplementedError:
          response =  self.error(NOFUNC)
          return response.asResponse()
      except Exception, e:
        Log.logException(e)
        # restart worker
        local.process_waiting.cancel()
        worker = pool.getWorkerObject(worker) # XXX shouldn't we time and control this thing here?
        # restart timing
        local.process_waiting = threading.Timer(config.instance_timeout, self.instanceRestart, kwargs={'target':this, 'worker':worker})
        try:
          data = self._process(worker, funcname, request)
        except Exception, e:
          # rationale: if the same happened twice, it was not the worker's fault, apparently the doc is lame
          Log.logException(e)
          local.process_waiting.cancel()
          response = self.error(BAD_DOCUMENT)
          return response.asResponse()
    finally:
      # release semaphore only if we were able to release worker
      if pool.releaseWorkerObject(worker):
        self.semaphore.release()
    local.process_waiting.cancel()
    if this.timed_out: # means it took far too long and we have already triggered an instance restart
      response = self.error(TIMEOUT) # XXX why not return data, after all? To inform the end-user that something is wrong.
      return response.asResponse()
    # remove None's
    if type(data) == type({}):
      for k, v in data.items():
        if v is None:
          data[k] = -1
    if type(data) in (type(()), type([])):
      data = [v is None and -1 or v for v in data]
    if data is None:
      data = -1
    response = oood_common.Response(OK, data)
    return response.asResponse()

  def instanceRestart(self, target, worker):
    """
      Restarts one worker instance, and makes it available
      To be triggered by timer threads in case thee was a timeout
    """
    Log.debug('instanceRestart: target %s, worker %s' % (str(target), str(worker)))
    # mark processing thread as timed_out in case it goes on
    target.timed_out = True
    # kill and restart worker
    worker = pool.getWorkerObject(worker)
    # if successful:
    if worker: # free worker and if ok then release semaphore
      if pool.releaseWorkerObject(worker):
        self.semaphore.release()
    else:
      Log.error('can not produce a new worker')
    # else: worker can not be restarted, pool becomes smaller, if it goes down to zero then next request
    # triggers reinitialization

  def emergencyRestart(self):
    """
      Restart of the whole machinery if for some reason
      the semaphore can not be acquired (because it is fully booked
      and there is nobody to release it)
    """
    Log.error('emergencyRestart triggered')
    # we want to do it only once, subsequent threads can return immediately
    if not self.emergency_restart_lock.acquire(False):
      Log.debug('somebody is doing emergencyRestart already')
      return
    # tell threads we are blocked
    for thr in threading.enumerate():
      thr.oood_has_blocked = True
    self.oood_has_blocked = True
    # release all semaphores so that pending threads can return
    for i in range(config.pool_size):
      self.semaphore.release()
    self.initialize()

  def error(self, code, message=''):
    Log.info(code)
    return oood_common.Response(code, message=message)

  def test(self, worker, *a, **kw):
    return 'tested with %s, %s and %s' % (worker, a, kw)

  def threadingStatus(self):
    s = ''
    line = '%-20s|%-20s|%-20s|%-20s|\n' % ('name', 'function', 'args', 'kwargs')
    s += line
    line = '%s%s%s%s\n' % tuple([20*'_'+'+' for i in range(4)])
    s += line
    for thr in threading.enumerate():
      line = ''
      line += '%-20s|' % thr.getName()
      isTimer = hasattr(thr, 'function')
      fname = isTimer and thr.function.__name__ or ''
      fargs = isTimer and thr.args or ''
      fkwargs   = isTimer and thr.kwargs or ''
      line += '%-20s|' % fname
      line += '%-20s|' % fargs
      line += '%-20s|' % fkwargs
      line += '\n'
      s += line
    return s

  def poolStatus(self):
    current_time = int(time.time())
    s = ''
    s += 'Server up: %s\n' % formatTime(current_time - self.creation_time)
    s += 'Restarts: %s\n' % self.restart_count
    s += 'Up since last restart: %s\n' % formatTime(current_time - self.start_time)
    s += 'Processed requests: %s\n' % self.counter
    s += '\n'
    line = '%-5s|%-8s|%-5s|%-5s|%-30s|\n' % ('idx', 'up', 'count', 'busy', 'file')
    s += line
    line = '%s+%s+%s+%s+%s\n' % ('-'*5, '-'*8, '-'*5, '-'*5, '-'*30)
    s += line
    worker_index_list = pool.inst.keys()
    worker_index_list.sort()
    for idx in worker_index_list:
      worker = pool.inst[idx]
      if worker is None:
        up = count = busy = file = 'xxxx'
      else:
        busy = worker.busytime or '-'
        up = formatTime(worker.uptime)
        count = worker.counter
        file = getattr(worker, 'fileUrl', None) or '-'
      line = '%-5s|%-8s|%-5s|%-5s|%-30s|\n' % (idx, up, count, busy, file)
      s += line
    return s

  def _process(self, worker, funcname, request):
    """
      processing function - builds an appropriate function name, calls it on the worker
      then processes return value
      kw is a Request instance (a dict-like object), the name is historical and will be
      changed in the next cleanup.
    """
    kw = request
    #argtpl=('filename', 'data', 'meta', 'extension')
    #kw = dict(zip(argtpl, a)) # build a dict from positional arguments
    method = getattr(worker, 'run_'+funcname, None)
    if method is None:
      method = getattr(worker, funcname, None) # method may have no prefix (we'll stop it some day anyway)
      if method is None:
        raise NotImplementedError('no function named ' + funcname)
    # if data given, store it in a temp file and proceed using std interface
    if kw.get('data'):
      if not kw.get('filename'):
        kw['filename'] = str(random.random()).split('.')[1]
      # we have to store in a file for OOo to open
      filename = '%d_%s' % (worker.idx, lib.asciify(kw['filename'])) # prepend worker index in case we have two files of the same name
      filename = self._mkName(filename)
      f = open(filename, 'wb')
      try:
        f.write(kw['data'])
      finally:
        f.close()
      kw['filename'] = filename
      kw['data'] = '' # not needed now, we have decoded data in the file
      try:
        # DO THE JOB
        res = method(kw) # res is not used, method puts processed data back into kw
        if kw.get('newfilename'): # worker produced data and saved it in a new file
          # XXX here we will also zip html files
          try: # we can do it only if the argument is a string
            tozip = kw.get('extension').endswith('html') # if html, we pack the output files
          except AttributeError:
            tozip = False
          if tozip:
            self._zipResult(kw)
          else:
            f = open(self._mkName(kw['newfilename']), 'rb')
            try:
              kw['data'] = f.read()
            finally:
              f.close()
        self.counter += 1
      finally:
        # remove temporary files
        if kw.get('filename') is not None:
          self._safeRemove(kw['filename'])
        if kw.get('newfilename') is not None:
          self._safeRemove(kw['newfilename'])
      return kw
    else:
      self.counter += 1
      return method(request)

  def _zipResult(self, kw):
    #first delete the original
    fname = kw['filename']
    fullname = self._mkName(fname)
    os.remove(fullname)
    arch = cStringIO.StringIO()
    pack = zipfile.ZipFile(arch, 'a')
    r = re.compile(kw['extension'] + '$')
    for f in glob.glob('%s/*'%self.tmp_dir_name):
      if os.path.isdir(f):
        continue
      name = os.path.basename(f)
      pack.write(f, name)
    pack.close()
    arch.seek(0)
    kw['data'] = arch.read()
    arch.close()
    kw['mime'] = 'application/zip' # generation returns mime type
    kw['newfilename'] += '.zip'

  def _safeRemove(self, filename):
    """
      remove a file or log warning if none exists
    """
    if config.debug_mode:
      return
    any = False
    pth = os.path.join(self.tmp_dir_name, os.path.basename(filename))
    for f in glob.glob('%s/*'%self.tmp_dir_name):
      if not config.debug_mode:
        if os.path.isdir(f):
          shutil.rmtree(f)
        else:
          os.remove(f)
      any = True
    #if not any:
      #Log.warning('no file beginning with %s' % filename)

  def _mkName(self, fname):
    """
      make a complete file name with path
    """
    return os.path.join(self.tmp_dir_name, fname)

if __name__=='__main__':
    ser = MySerw((config.server_host, config.server_port), allow_none = True)
    disp = Dispatcher()
    ser.register_instance(disp)
    ser.serve_forever()




# vim: filetype=python syntax=python shiftwidth=2 
