#!/usr/bin/python

"""
  This module is used by the server, and can be used by client
  to make life easier.
  It provides some utility functions, status code definitions and more.
"""

import base64
import re
import string
import types

nulltrans = string.maketrans('', '')
remother = nulltrans.translate(nulltrans, string.letters + string.digits + '.')

def asciify(s):
  """
    for clearing chars out of filenames
  """
  if type(s) == types.UnicodeType:
    return s.encode('ASCII', 'ignore')
  return s.translate(nulltrans, remother)


class MalformedRequest(Exception): pass
class InvalidStatusCode(Exception): pass

malformed_request_message = 'oood accepts either 1-4 arguments: filename [, data [, meta [, extension]] or only one argument - a dictionary'


code_map = {
  200: 'OK',
  401: 'requested method is not provided',
  402: 'the document could not be processed',
  403: 'timeout while processing document',
  404: 'malformed request',
  501: 'the server is restarting',
  502: 'pool is empty, the server will be restarted',
  503: 'the server is out of sync and will be restarted',
  504: 'unknown error'
  }


def code2message(code):
  """
    Returns status message of the given status code
  """
  return code_map.get(int(code), 'unknown code')


class Response(object):
  """
    This class represents a response give by oood.
    It is used by oood to produce a response.
    Can be used by client - an instance of this class should
    in this case be produced by responseFactory function.
    Response is given as a tuple of three items:
    - status code
    - response data
    - message
    Response data is a dictionary, which either has multiple keys (filename, data encoded in base64 etc),
    or only one key 'response_data'
  """

  code = 0

  def __init__(self, code, data={}, message='', sending=True):
    self.code = code
    # data is always a dict - if return value is not a dict, we
    # make a dict {'response_data': data}
    if type(data) == types.DictType:
      self.data = data
    else:
      self.data = dict(response_data=data)
    # if we have data we have to encode it (if we are sending)
    # or decode (if we are receiving)
    if self.data.has_key('data'):
      if sending:
        self.data['data'] = base64.encodestring(self.data['data'])
      else:
        self.data['data'] = base64.decodestring(self.data['data'])
    if not sending:
      self.__dict__.update(self.data) # to make returned data directly accessible
    self.message = message

  def asResponse(self):
    """
      Produces an appropriately formatted response.
    """
    return (self.code, self.data, self.message)

  def isCodeValid(func):
    """
      Decorator for checking if response code of self is valid.
      If it is not, an InvalidStatusCode exception is raised.
    """
    def wrappedFunc(self, *args, **kwargs):
      if self.code not in code_map.keys():
        raise InvalidStatusCode(str(self.code))
      return func(self, *args, **kwargs)
    return wrappedFunc

  def isOk(self):
    """
      Check if the response is "OK".
    """
    return self.code == 200
  isOk = isCodeValid(isOk)

  def isError(self):
    """
      Check if the response signals an error.
    """
    return self.code != 200
  isError = isCodeValid(isError)

  def isClientError(self):
    """
      Check if the error was caused by client, or was it an
      internal problem of the server.
    """
    return self.code > 400 and self.code < 500
  isClientError = isCodeValid(isClientError)

  def isServerError(self):
    """
      Check if the error was caused by client, or was it an
      internal problem of the server.
    """
    return self.code > 500
  isServerError = isCodeValid(isServerError)

def responseFactory(response):
  """
    To be used by clients.
    Given a tuple received from the server, produces an instance
    of Response class.
  """
  return Response(response[0], response[1], response[2], False)


class Request:
  """
    This class represents a request. It is to be used by the client
    to build a request and send is as the first (and only) argument; the 
    ServerProxy will convert it into a dictionary.
  """

  def __init__(self, **kw):
    for name, value in kw.items():
      self.__dict__[name] = self.process(name, value)

  def process(self, name, value):
    if name == 'data': # encode data
      value = base64.encodestring(value)
    if name == 'filename': # clear file name from non-ascii chars
      value = asciify(value)
    if name == 'encoding': # translate encoding to OOo-compatible syntax
      matches = (
          ('iso.?(\d{4}).?(\d)', 'ISO_%s_%s'),
          ('utf.?(\d+)', 'UTF%s'),
          )
      for pattern, template in matches:
        match = re.match(pattern, value, re.IGNORECASE)
        if match:
          return  template % match.groups()
    return value

  def __setattr__(self, name, value):
    """
      Encode binary data on the fly
    """
    self.__dict__[name] = self.process(name, value)

  def __getattr__(self, name):
    """
      Decode binary data on the fly
    """
    if name == 'data':
      return base64.decodestring(self.__dict__['data'])
    else:
      return self.__dict__[name]


def requestFactory(a):
  """
    To be used by server.
    Makes a Request object (a dictionary, for the time being), constructed from what was provided:
    if there was multiple various arguments, makes a dictionary using hardcoded names (old implementation).
    If it was a dictionary, do nothing.
    Decode base64 data if any were provided.
  """
  # fail if bad request
  if len(a) == 0: raise MalformedRequest(malformed_request_message)
  if type(a[0]) not in (types.StringType, types.UnicodeType, types.DictType):
    raise MalformedRequest(malformed_request_message)
  if type(a[0]) == types.DictType and len(a) > 1: raise MalformedRequest(malformed_request_message)
  # process
  if type(a[0]) in (types.StringType, types.UnicodeType):
    argtpl=('filename', 'data', 'meta', 'extension', 'orig_format')
    kw = dict(zip(argtpl, a)) # build a dict from positional arguments
  if type(a[0]) == types.DictType:
    kw = a[0]
  # XXX this does not work as expected (yet), so we use a plain dictionary and decode data here
  # we will be usint Request class later
  if kw.has_key('data'):
    kw['data'] = base64.decodestring(kw['data'])
  return kw
  

if __name__=='__main__':
  r = Request(a=123, data='adsfadf')
  print r.a
  #r.data = 'asdfasdf'
  print r
  print r.data
  print r['data']
  print r.get('data')
  print r.has_key('data')
